import java.io.*;
import javax.swing.*;
import java.awt.event.*;
import javax.swing.filechooser.*;
import java.util.*;
import java.util.stream.Collectors;

class Node {
	char item;
	int frequency;
	Node parent;
	ArrayList<Node> children;
	Node nextSameItem;

	Node() {
		item = '%';
		frequency = -1;
		parent = null;
		children = new ArrayList<Node>();
		nextSameItem = null;
	}

	Node(char p) {
		item = p;
		frequency = 1;
		parent = null;
		children = new ArrayList<Node>();
		nextSameItem = null;
	}

	void addchild(char p) {
		children.add(new Node(p));
		children.get(children.size() - 1).parent = this;
	}

	Node findchild(char p) {
		for (Node child : children) {
			if (child.item == p)
				return child;
		}
		return null;
	}
}

class FpTree {

public String frequent_items_string;

FpTree(){frequent_items_string="";}


	public ArrayList<Node> prepareheader(ArrayList<ArrayList<Character>> transactions) {
		HashMap<Character, Integer> counterMap = new HashMap<Character, Integer>();
		ArrayList<Character> ignoreList = new ArrayList<Character>();
		for (ArrayList<Character> transaction : transactions) {
			for (int i = 0; i < transaction.size(); i++) {
				Character product = (Character) transaction.get(i);
				counterMap.put(product, counterMap.getOrDefault(product, 0) + 1);
			}
		}

		ArrayList<Node> fpHeader = new ArrayList<Node>();
		for (Character product : counterMap.keySet()) {
			if (counterMap.get(product) >= 3)
				fpHeader.add(new Node(product));
			else
				ignoreList.add(product);
		}
		fpHeader.sort((t1, t2) -> counterMap.get(t2.item) - counterMap.get(t1.item));
		for (ArrayList<Character> transaction : transactions) {
			transaction.removeAll(ignoreList);
			transaction.sort((t1, t2) -> counterMap.get(t2) - counterMap.get(t1));
		}

		return fpHeader;
	}

	public static Node getheader(ArrayList<Node> headertable, char item) {
		for (Node product : headertable) {
			if (product.item == item)
				return product;
		}
		return null;
	}

	public Node constructFPTree(ArrayList<ArrayList<Character>> transactions, ArrayList<Node> fpHeader) {
		Node root = new Node();
		for (ArrayList<Character> transaction : transactions) {
			Node temp = root;
			for (int index = 0; index < transaction.size(); index++) {
				Character product = (Character) transaction.get(index);
				Node child = temp.findchild(product);
				if (child == null) {
					temp.addchild(product);
					temp = temp.children.get(temp.children.size() - 1);
					Node next = getheader(fpHeader, product);
					while (next.nextSameItem != null)
						next = next.nextSameItem;
					next.nextSameItem = temp;
				} else {
					child.frequency += 1;
					temp = child;
				}

			}
		}
		return root;
	}

	public void printFPTree(Node root)
	{
		frequent_items_string += "FP-TREE"+"\nroot";
		Queue<Node> q = new LinkedList<Node>();
		q.add(root);
		q.add(new Node('&'));
		while(q.size()!=0)
		{
			Node printnode = q.remove();	
			if(printnode.item == '&')
				{if(q.size()==0)
					{break;}
				q.add(new Node('&'));
				frequent_items_string += "\n" ;}
			else
			{
				if(printnode.item != '%')
				{
				frequent_items_string += printnode.item + ":" + printnode.frequency + "(" + printnode.parent.item + ")" + "      ";
				}
				for(Node temp : printnode.children)
					{q.add(temp);}
			}
		}
		frequent_items_string += "\n\n"+"FREQUENT ITEMSETS\n";
	}

	public void fpGrowth(ArrayList<ArrayList<Character>> transRecords, ArrayList<Character> postPattern) {
		ArrayList<Node> trees = prepareheader(transRecords);
		Node root1 = constructFPTree(transRecords, trees);
		if (root1.children.size() == 0) {
			return;
		}
		if (postPattern.size() > 0) {
			for (Node node : trees) {
				frequent_items_string += node.item;
				for (char val : postPattern) {
					frequent_items_string += " " + val;
				}
				frequent_items_string += "\n";
			}
		}
		for (int i = trees.size() - 1; i >= 0; i--) {
			Node node = trees.get(i);
			ArrayList<Character> tmp = new ArrayList<>();
			tmp.add(node.item);
			if (postPattern.size() > 0) {
				tmp.addAll(postPattern);
			}
			ArrayList<ArrayList<Character>> records = new ArrayList<>();
			Node nextNode = node.nextSameItem;
			while (nextNode != null) {
				int frequency = nextNode.frequency;
				ArrayList<Character> preNodes = new ArrayList<Character>();
				Node parent = nextNode;
				while ((parent = parent.parent) != null && parent.item != '%') {
					preNodes.add(parent.item);
				}
				while (frequency > 0) {
					frequency--;
					records.add(preNodes);
				}
				nextNode = nextNode.nextSameItem;
			}
			fpGrowth(records, tmp);
		}
	}



	public void getInputDataFromFile(String filename, ArrayList<ArrayList<Character>> input) {
		FileReader fileReader = null;
		Scanner inputFile = null;
		try {
			fileReader = new FileReader(filename);
			inputFile = new Scanner(fileReader);
			String line;
			while (inputFile.hasNextLine()) {
				line = inputFile.nextLine().trim();
				ArrayList<Character> strArr = (ArrayList<Character>) Arrays.stream(line.split(","))
						.map(c -> c.charAt(0)).collect(Collectors.toList());
				input.add(strArr);
			}
		} catch (IOException e) {
			System.err.println(e);
		} finally {
			inputFile.close();
		}
	}
}



class MyProject extends JFrame implements ActionListener {
 
    static JLabel l;
 
  
    MyProject()
    {
    }
 
    public static void main(String args[])
    {
        JFrame f = new JFrame("FREQUENT PATTERN MINING");

        f.setSize(1000, 1000);
        f.setVisible(true); 
        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        JButton fileselect = new JButton("SELECT FILE");
        JButton runalgorithm = new JButton("SUBMIT");
 
     
        MyProject f1 = new MyProject();

        fileselect.addActionListener(f1);
        runalgorithm.addActionListener(f1);

        JPanel p = new JPanel();
 
        p.add(fileselect);
        p.add(runalgorithm);
 
        l = new JLabel("NO FILE SELECTED");
 
        p.add(l);
        f.add(p);
 
        f.show();
    }

    public void actionPerformed(ActionEvent evt)
    {
        String command = evt.getActionCommand();
 
        if (command.equals("SUBMIT")) {
            if(l.getText()!= "NO FILE SELECTED")
		 {
			ArrayList<ArrayList<Character>> input = new ArrayList<>();
		
		try{
		FpTree fptree = new FpTree();
		fptree.getInputDataFromFile(l.getText(), input);
		ArrayList<Character> in = new ArrayList<Character>();
		fptree.printFPTree(fptree.constructFPTree(input,fptree.prepareheader(input)));
		fptree.fpGrowth(input, in);
		FileWriter fw = new FileWriter("output_fptree_itemsets.txt");
          	for(int i=0;i<fptree.frequent_items_string.length(); i++)
               				{fw.write(fptree.frequent_items_string.charAt(i));}
  			
		 l.setText(l.getText()+"  successfully mined fp tree and found frequent itemsets");
            System.out.println("Successfully written");
		
            fw.close();
		}catch(Exception e){}

		 }
		 else
		 {
			l.setText("PLEASE SELECT INPUT FILE");
		 }
       }
 
     
        else {
           
            JFileChooser j = new JFileChooser(FileSystemView.getFileSystemView().getHomeDirectory());
 
            int r = j.showOpenDialog(null);
            if (r == JFileChooser.APPROVE_OPTION)
            {
             l.setText(j.getSelectedFile().getAbsolutePath());
            }
            else
                l.setText("NO FILE SELECTED");
        }
    }
}